# Copyright (c) 2022 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module containing the entry point for the ros_license_toolkit CLI.
"""

import argparse
import os
import sys
import timeit
from typing import Optional, Sequence

from ros_license_toolkit.checks import Status
from ros_license_toolkit.license_checks.license_file_referenced_check import (
    LicenseFilesReferencedCheck,
)
from ros_license_toolkit.license_checks.license_in_code_check import LicensesInCodeCheck
from ros_license_toolkit.license_checks.license_tag_exists_check import LicenseTagExistsCheck
from ros_license_toolkit.license_checks.license_tag_is_spdx import LicenseTagIsInSpdxListCheck
from ros_license_toolkit.license_checks.license_text_exists_check import LicenseTextExistsCheck
from ros_license_toolkit.license_checks.schema_check import SchemaCheck
from ros_license_toolkit.package import get_packages_in_path
from ros_license_toolkit.ui_elements import (
    FAILURE_STR,
    SUCCESS_STR,
    WARNING_STR,
    Verbosity,
    major_sep,
    minor_sep,
    red,
    rll_print_factory,
)


def main(args: Optional[Sequence[str]] = None) -> int:
    """Main entry point for the ros_license_toolkit CLI.

    :param args: the command line arguments, defaults to sys.argv[1:]
    :type args: Sequence[str], optional
    :return: the exit code, `os.EX_OK` if all checks passed, `os.EX_USAGE` if
        no packages were found, `os.EX_DATAERR` if any check failed
    :rtype: int
    """
    if not args:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description="Checks ROS packages for correct license declaration."
    )
    parser.add_argument(
        "path", default=".", help="path to ROS2 package or repo containing packages"
    )
    parser.add_argument(
        "-c",
        "--generate_copyright_file",
        action="store_true",
        default=False,
        help="generate a copyright file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="enable verbose output",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=False,
        help="disable most output",
    )
    parser.add_argument(
        "-e",
        "--continue_on_error",
        action="store_true",
        default=False,
        help="treats all errors as warnings, i.e. will give " + "returncode 0 even on errors",
    )
    parser.add_argument(
        "-w",
        "--warnings_as_error",
        action="store_true",
        default=False,
        help="treats all warnings as errors",
    )
    parsed_args = parser.parse_args(args)

    # Determine the verbosity level
    if parsed_args.quiet:
        verbosity = Verbosity.QUIET
    elif parsed_args.verbose:
        verbosity = Verbosity.VERBOSE
    else:
        verbosity = Verbosity.NORMAL
    rll_print = rll_print_factory(verbosity)

    # Sanity check the path
    assert isinstance(parsed_args.path, str)
    assert os.path.exists(parsed_args.path), f"Path {parsed_args.path} does not exist."

    # Get the packages in the path
    packages = get_packages_in_path(parsed_args.path)
    if not packages:
        rll_print(f"No packages found in {parsed_args.path}", Verbosity.QUIET)
        return os.EX_USAGE
    rll_print(f"Found {len(packages)} packages in {parsed_args.path}", Verbosity.QUIET)
    rll_print(f' Packages: {", ".join([p.name for p in packages])}', Verbosity.VERBOSE)
    rll_print(major_sep())

    # lets time the execution
    start = timeit.default_timer()

    # Check the packages
    results_per_package = {}
    for package in packages:
        results_per_package.update(process_one_pkg(rll_print, package))

    if parsed_args.generate_copyright_file:
        if max(results_per_package.values()) != Status.FAILURE:
            generate_copyright_file(packages, rll_print)
        else:
            rll_print(
                red("Copyright file will not be generated " + "because there were linter errors."),
                Verbosity.QUIET,
            )

    stop = timeit.default_timer()
    rll_print(f"Execution time: {stop - start:.2f} seconds", Verbosity.QUIET)

    # Print the overall results
    return print_results(results_per_package, rll_print, parsed_args)


def generate_copyright_file(packages, rll_print):
    """Generate copyright file. In case more than one package
    is provided, display error message."""
    if len(packages) == 1:
        package = packages[0]
        try:
            package.write_copyright_file(os.path.join(os.getcwd(), "copyright"))
        except AssertionError as error:
            rll_print(red(str(error)))
    else:
        rll_print(red("Can only generate copyright file for single package"), Verbosity.QUIET)


def print_results(result, rll_print, args):
    """Printing the result of package"""
    if max(result.values()) == Status.SUCCESS:
        rll_print(f"All packages:\n {SUCCESS_STR}", Verbosity.QUIET)
        return os.EX_OK

    if max(result.values()) == Status.WARNING:
        if args.warnings_as_error:
            rll_print(
                f"All packages:\n {FAILURE_STR} " + "(Treating warnings as failure)",
                Verbosity.QUIET,
            )
            return os.EX_DATAERR
        rll_print(f"All packages:\n {WARNING_STR}", Verbosity.QUIET)
        return os.EX_OK

    if args.continue_on_error:  # Error is Warning, still displayed red
        rll_print(
            f"All packages:\n {WARNING_STR} " + "(Treating errors as warnings)",
            Verbosity.QUIET,
        )
        return os.EX_OK
    rll_print(f"All packages:\n {FAILURE_STR}", Verbosity.QUIET)
    return os.EX_DATAERR


def process_one_pkg(rll_print, package):
    """Perform checks on one package, print results and return them."""
    results_per_package = {}
    rll_print(f"[{package.name}]")
    assert package.repo is not None, "Package must be in a git repo."
    rll_print(f"git hash of ({package.repo.get_path()}):" f" {package.repo.get_hash()}")
    checks_to_perform = [
        SchemaCheck(),
        LicenseTagExistsCheck(),
        LicenseTagIsInSpdxListCheck(),
        LicenseTextExistsCheck(),
        LicensesInCodeCheck(),
        LicenseFilesReferencedCheck(),
    ]

    for check in checks_to_perform:
        check.check(package)
        rll_print(check)
        rll_print(check.verbose(), Verbosity.VERBOSE)

    rll_print(minor_sep())
    # Every check is successful, no warning
    if max(check.status for check in checks_to_perform) == Status.SUCCESS:
        rll_print(f"[{package.name}] Overall:\n {SUCCESS_STR}")
        results_per_package[package.abspath] = Status.SUCCESS
    # Either every check is successful or contains a warning
    elif max(check.status for check in checks_to_perform) == Status.WARNING:
        rll_print(f"[{package.name}] Overall:\n {WARNING_STR}")
        results_per_package[package.abspath] = Status.WARNING
    # At least one check contains an error
    else:
        rll_print(f"[{package.name}] Overall:\n {FAILURE_STR}")
        results_per_package[package.abspath] = Status.FAILURE
    rll_print(major_sep())
    return results_per_package


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains LicensesInCodeCheck, which implements Check"""

from pprint import pformat
from typing import Dict, List

from ros_license_toolkit.checks import Check
from ros_license_toolkit.license_tag import LicenseTag
from ros_license_toolkit.package import Package
from ros_license_toolkit.ui_elements import red


class LicensesInCodeCheck(Check):
    """Check if all found licenses have a declaration in the package.xml."""

    def __init__(self: "LicensesInCodeCheck"):
        Check.__init__(self)
        self.declared_licenses: Dict[str, LicenseTag] = {}
        self.files_with_uncovered_licenses: Dict[str, List[str]] = {}
        self.files_not_matched_by_any_license_tag: Dict[str, List[str]] = {}
        self.files_with_unofficial_tag: Dict[str, List[str]] = {}

    def _check(self, package: Package):
        if len(package.license_tags) == 0:
            self._failed("No license tag defined.")
            return
        self.declared_licenses = package.license_tags
        self._check_license_files(package)
        self._evaluate_result(package)

    def _check_license_files(self, package: Package) -> None:
        for fname, found_licenses in package.found_files_w_licenses.items():
            if fname in package.get_license_files():
                # the actual license text files are not relevant for this
                continue
            found_licenses_str = found_licenses["detected_license_expression_spdx"]
            if not found_licenses_str:
                continue
            licenses = found_licenses_str.split(" AND ")
            for license_str in licenses:
                if license_str not in self.declared_licenses:
                    # this license has an unofficial tag
                    unofficial_licenses = {
                        lic_tag.id_from_license_text: key
                        for key, lic_tag in package.license_tags.items()
                        if lic_tag.id_from_license_text != ""
                    }
                    if license_str in unofficial_licenses.keys():
                        if fname not in self.files_with_unofficial_tag:
                            self.files_with_unofficial_tag[fname] = []
                        self.files_with_unofficial_tag[fname].append(license_str)
                        self.files_with_unofficial_tag[fname].append(
                            unofficial_licenses[license_str]
                        )
                        continue
                    # this license is not declared by any license tag
                    if fname not in self.files_with_uncovered_licenses:
                        self.files_with_uncovered_licenses[fname] = []
                    self.files_with_uncovered_licenses[fname].append(license_str)
                    continue
                if fname not in self.declared_licenses[license_str].source_files:
                    # this license is declared by a license tag but the file
                    # is not listed in the source files of the license tag
                    if fname not in self.files_not_matched_by_any_license_tag:
                        self.files_not_matched_by_any_license_tag[fname] = []
                    self.files_not_matched_by_any_license_tag[fname].append(license_str)
                    continue

    def _evaluate_result(self, package: Package) -> None:
        if self.files_with_uncovered_licenses:
            info_str = ""
            info_str += (
                "\nThe following files contain licenses that "
                + "are not covered by any license tag:\n"
                + "\n".join(
                    [f"  '{x[0]}': {x[1]}" for x in self.files_with_uncovered_licenses.items()]
                )
            )
            self._print_info(
                info_str,
                self.files_with_uncovered_licenses,
                self.files_not_matched_by_any_license_tag,
                package,
            )
        elif self.files_with_unofficial_tag:
            info_str = ""
            info_str += (
                "For the following files, please change the "
                + "License Tag in the package file to SPDX format:\n"
                + "\n".join(
                    [
                        f"  '{x[0]}' is of {x[1][0]} but its Tag is {x[1][1]}."
                        for x in self.files_with_unofficial_tag.items()
                    ]
                )
            )
            self._warning(info_str)
        elif len(self.files_not_matched_by_any_license_tag) > 0:
            info_str = ""
            info_str += (
                "\nThe following files contain licenses that "
                + "are covered by a license tag but are not listed in "
                + "the source files of the license tag:\n"
                + "\n".join(
                    [
                        f"  '{x[0]}': {x[1]}"
                        for x in self.files_not_matched_by_any_license_tag.items()
                    ]
                )
            )
            self._print_info(
                info_str,
                self.files_with_uncovered_licenses,
                self.files_not_matched_by_any_license_tag,
                package,
            )
        else:
            self._success("All licenses found in the code are covered by a license declaration.")

    def _print_info(
        self,
        info_str,
        files_with_uncovered_licenses,
        files_not_matched_by_any_license_tag,
        package,
    ):
        assert info_str != ""
        self._failed(info_str)
        self.verbose_output = red(
            "\n  Relevant scan results:\n"
            + pformat(
                list(
                    filter(
                        lambda x: x[0] in files_with_uncovered_licenses
                        or (x[0] in files_not_matched_by_any_license_tag),
                        package.found_files_w_licenses.items(),
                    )
                )
            )
        )



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains LicenseTagExistsCheck, which implements Check."""

from ros_license_toolkit.checks import Check
from ros_license_toolkit.package import Package
from ros_license_toolkit.ui_elements import red


class LicenseTagExistsCheck(Check):
    """This ensures that a tag defining the license exists."""

    def _check(self, package: Package):
        if len(package.license_tags) == 0:
            self._failed("No license tag defined.")
            self.verbose_output = red(str(package.package_xml))
        else:
            self._success(f"Found licenses {list(map(str, package.license_tags))}")



# Copyright (c) 2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utility functions."""

import os
from typing import Any, Dict, List, Optional

REQUIRED_PERCENTAGE_OF_LICENSE_TEXT = 90.0

# files we ignore in scan results
IGNORED = [
    "CHANGELOG.rst",
    ".scanignore",
    "package.xml",
    "setup.py",
    "setup.cfg",
    "CMakeLists.txt",
    ".git/*",
]


def get_spdx_license_name(scan_results: Dict[str, Any]) -> Optional[str]:
    """Get the SPDX license name from scan results."""
    if scan_results["percentage_of_license_text"] >= REQUIRED_PERCENTAGE_OF_LICENSE_TEXT:
        return scan_results["detected_license_expression_spdx"]
    return None


def get_ignored_content(pkg_abspath: str) -> List[str]:
    """Return all ignored patterns from '.scanignore'
    and local IGNORED definition."""
    ignored_content: List[str] = []
    scanignore_path = pkg_abspath + "/.scanignore"
    if os.path.exists(scanignore_path):
        with open(scanignore_path, "r", encoding="utf-8") as f:
            for line in f:
                line_contents = line.split("#")
                ignore_pattern = line_contents[0].rstrip()
                if len(ignore_pattern) > 0:
                    ignored_content.append(ignore_pattern)
            f.close()
    for pattern in IGNORED:
        if pattern not in ignored_content:
            ignored_content.append(pattern)
    return ignored_content



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains LicenseTextExistsCheck, which implements Check."""

import os
from typing import Any, Dict, Optional

import requests  # type: ignore[import-untyped]
from Levenshtein import distance

from ros_license_toolkit.checks import Check, Status
from ros_license_toolkit.common import get_spdx_license_name
from ros_license_toolkit.license_tag import LicenseTag, is_license_name_in_spdx_list
from ros_license_toolkit.package import Package
from ros_license_toolkit.ui_elements import red

# Value for minimal percentage between license texts for them to be accepted
SIMILARITY_THRESHOLD = 90  # in percent


class LicenseTextExistsCheck(Check):
    """This ensures that the license text file referenced by the tag exists."""

    def __init__(self: "LicenseTextExistsCheck"):
        Check.__init__(self)
        self.license_tags_without_license_text: Dict[LicenseTag, str] = {}
        self.missing_license_texts_status: Dict[LicenseTag, Status] = {}
        self.files_with_wrong_tags: Dict[LicenseTag, Dict[str, str]] = {}
        self.found_license_texts: Dict[str, Any] = {}

    def _check(self, package: Package):
        if len(package.license_tags) == 0:
            self._failed("No license tag defined.")
            return

        self._check_licenses(package)
        self._evaluate_results()

    def _check_licenses(self, package: Package) -> None:
        """checks each license tag for the corresponding license text. Also
        detects unofficial licenses when tag is not in the SPDX license list"""
        self.found_license_texts = package.found_license_texts
        for license_tag in package.license_tags.values():
            if not license_tag.has_license_text_file():
                self.license_tags_without_license_text[license_tag] = (
                    "No license text file defined."
                )
                self.missing_license_texts_status[license_tag] = Status.FAILURE
                continue
            license_text_file = license_tag.get_license_text_file()
            if not os.path.exists(os.path.join(package.abspath, license_text_file)):
                self.license_tags_without_license_text[license_tag] = (
                    f"License text file '{license_text_file}' does not exist."
                )
                self.missing_license_texts_status[license_tag] = Status.FAILURE
                continue
            if license_text_file not in self.found_license_texts:
                self.license_tags_without_license_text[license_tag] = (
                    f"License text file '{license_text_file}' not included" + " in scan results."
                )
                self.missing_license_texts_status[license_tag] = Status.FAILURE
                continue
            if not get_spdx_license_name(self.found_license_texts[license_text_file]):
                self.license_tags_without_license_text[license_tag] = (
                    f"License text file '{license_text_file}' is not "
                    + "recognized as license text."
                )
                self.missing_license_texts_status[license_tag] = Status.FAILURE
                continue
            actual_license: Optional[str] = get_spdx_license_name(
                self.found_license_texts[license_text_file]
            )
            if actual_license is None:
                self.license_tags_without_license_text[license_tag] = (
                    f"License text file '{license_text_file}'"
                    + " is not recognized as license text."
                )
                self.missing_license_texts_status[license_tag] = Status.FAILURE
                continue

            if actual_license != license_tag.get_license_id():
                if license_tag.has_license_text_file():
                    license_file_for_tag = (
                        package.abspath + "/" + license_tag.get_license_text_file()
                    )
                with open(license_file_for_tag, "r", encoding="utf-8") as f:
                    content = f.read()
                    similarity_of_texts = self.compare_text_with_spdx_text(license_tag, content)

                # IDEA: if accepted, add the tag to the package.found_license_texts, since scanning
                # has failed to do so. Also solves problem of license_file_referenced check

                # if similarity couldn't be determined or is too low --> fail, else success
                if similarity_of_texts is None or similarity_of_texts < SIMILARITY_THRESHOLD:
                    self.license_tags_without_license_text[license_tag] = (
                        f"License text file '{license_text_file}' is "
                        + f"of license {actual_license} but tag is "
                        + f"{license_tag.get_license_id()}."
                    )
                    # If Tag and File both are in SPDX but don't match -> Error
                    if is_license_name_in_spdx_list(license_tag.get_license_id()):
                        self.missing_license_texts_status[license_tag] = Status.FAILURE
                    else:
                        self.missing_license_texts_status[license_tag] = Status.WARNING
                    self.files_with_wrong_tags[license_tag] = {
                        "actual_license": actual_license,
                        "license_tag": license_tag.get_license_id(),
                    }
                    continue

    def _evaluate_results(self):
        if len(self.license_tags_without_license_text) > 0:
            if max(self.missing_license_texts_status.values()) == Status.WARNING:
                self._warning(
                    "Since they are not in the SPDX list, we can not check if these tags have the"
                    " correct license text:\n"
                    + "\n".join(
                        [
                            f"  '{x[0]}': {x[1]}"
                            for x in self.license_tags_without_license_text.items()
                        ]
                    )
                )
            else:
                self._failed(
                    "The following license tags do not have a valid license text file:\n"
                    + "\n".join(
                        [
                            f"  '{x[0]}': {x[1]}"
                            for x in self.license_tags_without_license_text.items()
                        ]
                    )
                )
            self.verbose_output = red(  # pylint: disable=attribute-defined-outside-init
                "\n".join([f"  '{x[0]}': {x[1]}" for x in self.found_license_texts.items()])
            )
        else:
            self._success("All license tags have a valid license text file.")

    def compare_text_with_spdx_text(self, tag, found_lic_text):
        """Get similarity percent between original license text (from spdx api) and given license
        text."""
        cache_dir: str = os.path.expanduser("~/.cache/ros_license_toolkit")
        os.makedirs(cache_dir, exist_ok=True)
        license_file = os.path.join(cache_dir, f"license_{tag}.txt")

        if not os.path.exists(license_file):
            url = f"https://spdx.org/licenses/{tag}.json"
            response = requests.get(url, timeout=100)
            if response is not None and response.status_code == 200:
                parsed_response = response.json()
                original_text = parsed_response["licenseText"]
                with open(license_file, "w", encoding="utf-8") as f:
                    f.write(original_text)
            else:
                return None
        else:
            with open(license_file, "r", encoding="utf-8") as f:
                original_text = f.read()
        difference = self.get_similarity_percent(original_text, found_lic_text)
        return difference

    def get_similarity_percent(self, text1, text2):
        """Levenshtein distance based similarity percent of text1 and text2, regularized to longer
        text for percent value."""
        lev_dis = distance(text1, text2)
        bigger = max(len(text1), len(text2))
        similarity_percentage = round(100 * (bigger - lev_dis) / bigger, 2)
        return similarity_percentage



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains LicenseTagIsInSpdxListCheck, which implements Check."""

from ros_license_toolkit.checks import Check
from ros_license_toolkit.license_tag import is_license_name_in_spdx_list
from ros_license_toolkit.package import Package


class LicenseTagIsInSpdxListCheck(Check):
    """This ensures that the license tag is in the SPDX list of licenses."""

    def _check(self, package: Package):
        licenses_not_in_spdx_list = []
        for license_tag in package.license_tags.keys():
            if not is_license_name_in_spdx_list(license_tag):
                licenses_not_in_spdx_list.append(license_tag)
        if len(licenses_not_in_spdx_list) > 0:
            self._warning(
                f"Licenses {licenses_not_in_spdx_list} are "
                "not in SPDX list of licenses. "
                "Make sure to exactly match one of https://spdx.org/licenses/."
            )
        else:
            self._success("All license tags are in SPDX list of licenses.")



# Copyright (c) 2022 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the checks for the linter."""

from enum import IntEnum

from ros_license_toolkit.package import Package, PackageException
from ros_license_toolkit.ui_elements import NO_REASON_STR, green, red, yellow


class Status(IntEnum):
    """Levels of success or failure for the output"""

    SUCCESS = 0
    WARNING = 1
    FAILURE = 2


class Check:
    """Base class for checks."""

    def __init__(self: "Check"):
        """Initialize a check."""
        # overall success of this check
        self.status: Status = Status.FAILURE

        # explanation for success or failure for normal output
        self.reason: str = NO_REASON_STR

        # string with additional information for verbose output
        self.verbose_output: str = ""

    def _failed(self, reason: str):
        """Set this check as failed for `reason`."""
        self.status = Status.FAILURE
        self.reason = reason

    def _warning(self, reason: str):
        """Set this check as passed but display a warning for reason `r`."""
        self.status = Status.WARNING
        self.reason = reason

    def _success(self, reason: str):
        """Set this check as successful for reason `r`."""
        self.status = Status.SUCCESS
        if self.reason == NO_REASON_STR:
            self.reason = ""
        else:
            self.reason += "\n "
        self.reason += reason

    def __str__(self) -> str:
        """Return formatted string for normal output."""
        info: str = str(type(self).__name__) + "\n"
        if self.status == Status.SUCCESS:
            info += green(f" SUCCESS {self.reason}")
        elif self.status == Status.WARNING:
            info += yellow(f" WARNING {self.reason}")
        else:
            info += red(f" FAILURE {self.reason}")
        return info

    def verbose(self) -> str:
        """Return string with additional information for verbose output."""
        return self.verbose_output

    def __bool__(self) -> bool:
        """Evaluate success of check as bool. Warning is treated as success"""
        return self.status != Status.FAILURE

    def get_status(self) -> Status:
        """Get the level of success to also consider possible warnings"""
        return self.status

    def check(self, package: Package):
        """
        Check `package` and set success and reason.

        Calls `_check` internally. If they raise an exception, this is
        caught and the check is set to failed.
        """
        try:
            self._check(package)
        except PackageException as ex:
            self._failed(f"PackageException: {ex}")
            return
        except AssertionError as ex:
            self._failed(f"AssertionError: {ex}")
            return

    def _check(self, package: Package):
        """Check `package`. To be overwritten by subclasses."""
        raise NotImplementedError("Overwrite this")



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains LicenseFilesReferencedCheck, which implements Check"""

import os
from typing import Dict, List

from ros_license_toolkit.checks import Check
from ros_license_toolkit.package import Package


class LicenseFilesReferencedCheck(Check):
    """Check if all found License file have a reference in package.xml."""

    def __init__(self: "LicenseFilesReferencedCheck"):
        Check.__init__(self)
        self.not_covered_texts: Dict[str, str] = {}
        self.inofficial_covered_texts: Dict[str, List[str]] = {}

    def _check(self, package: Package):
        for filename, license_text in package.found_license_texts.items():
            # skipping all declarations above the package
            if not is_in_package(package, filename):
                continue
            self._handle_inofficial_licenses(package, filename, license_text)
        self._evaluate_results()

    def _handle_inofficial_licenses(self, package: Package, filename, license_text):
        if (
            "detected_license_expression_spdx" in license_text
            and license_text["detected_license_expression_spdx"] not in package.license_tags
        ):
            spdx_expression = license_text["detected_license_expression_spdx"]
            inofficial_licenses = {
                lic_tag.id_from_license_text: key
                for key, lic_tag in package.license_tags.items()
                if lic_tag.id_from_license_text != ""
            }
            if spdx_expression in inofficial_licenses:
                self.inofficial_covered_texts[filename] = [
                    spdx_expression,
                    inofficial_licenses[spdx_expression],
                ]
            else:
                self.not_covered_texts[filename] = spdx_expression

    def _evaluate_results(self):
        if self.not_covered_texts:
            info_str = ""
            info_str += (
                "The following license files are not"
                + " mentioned by any tag:\n"
                + "\n".join([f"  '{x[0]}' is of {x[1]}." for x in self.not_covered_texts.items()])
            )
            self._failed(info_str)
        elif self.inofficial_covered_texts:
            info_str = ""
            info_str += (
                "The following license files are not"
                + " mentioned by any tag:\n"
                + "\n".join(
                    [
                        f"  '{x[0]}' is of {x[1][0]} but its tag is {x[1][1]}."
                        for x in self.inofficial_covered_texts.items()
                    ]
                )
            )
            self._warning(info_str)
        else:
            self._success("All license declaration are referenced by a tag.")


def is_in_package(package: Package, file: str) -> bool:
    """Return TRUE if the file is underneath the absolute package path.
    Return FALSE if file is located above package."""
    parent = os.path.abspath(package.abspath)
    child = os.path.abspath(package.abspath + "/" + file)

    comm_parent = os.path.commonpath([parent])
    comm_child_parent = os.path.commonpath([parent, child])
    return comm_parent == comm_child_parent



# Copyright (c) 2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/ros_license_toolkit

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This Module contains SchemaCheck, which implements Check."""

import os
from typing import Optional, Tuple

import requests  # type: ignore[import-untyped]
from lxml import etree

from ros_license_toolkit.checks import Check
from ros_license_toolkit.package import Package


class SchemaCheck(Check):
    """This checks the xml scheme and returns the version number."""

    def __init__(self):
        super().__init__()
        self.accepted_versions = [1, 2, 3]
        self.validation_schema: Optional[etree.XMLSchema] = None

    def _check(self, package: Package):
        """Checks via scheme validation via self._validate.
        Also considers version of package.xml for validation."""
        version: int = package.package_xml_format_ver
        if version in self.accepted_versions:
            status, message = self.validate(package)
            if status:
                self._success(
                    f"Detected package.xml version {version}, " "validation of scheme successful."
                )
            else:
                reason = f"package.xml contains errors: {message}"
                self._failed(reason)
        else:
            # Temporary workaround for not implemented version 4
            if version == 4:
                reason = (
                    "couldn't check package.xml scheme. Version 4 is " + "not available right now"
                )
                self._warning(reason)
            else:
                reason = (
                    "package.xml does not contain correct package "
                    + "format number. Please use a real version. "
                    + '(e.g. <package format="3">)'
                )
                self._failed(reason)

    def validate(self, package: Package) -> Tuple[bool, str]:
        """This is validating the package.xml schema from given package.
        This can only validate for format version 1, 2 or 3. Every other
        version WILL FAIL. If everything is correct, returns format number,
        else -1."""
        version = package.package_xml_format_ver
        message = ""
        schema = self.get_validation_schema(version)
        if schema:
            result = schema.validate(package.parsed_package_xml)
            if not result:
                message = schema.error_log.last_error
            return result, message
        return False, "Couldn't get schema, no validation possible."

    def get_validation_schema(self, version: int):
        """Return validation schema for version 1, 2 or 3. If called for other
        version numbers, this WILL FAIL. Version is not checked again.
        Only call with version 1, 2 or 3."""
        cache_dir: str = os.path.expanduser("~/.cache/ros_license_toolkit")
        os.makedirs(cache_dir, exist_ok=True)
        schema_file = os.path.join(cache_dir, f"package_format{version}.xsd")

        if not os.path.exists(schema_file):
            address = f"http://download.ros.org/schema/package_format{version}.xsd"
            try:
                response = requests.get(address, stream=True, timeout=100)
                response.raise_for_status()
                response_parsed = response.content  # throw error when bad http code

                schema = etree.fromstring(response_parsed)
                with open(schema_file, "wb") as f:
                    f.write(etree.tostring(schema))
            except (AttributeError, etree.XMLSyntaxError) as error:
                print(error)
                print("An error encountered while getting " + address)
        else:
            schema = etree.parse(schema_file)

        self.validation_schema = etree.XMLSchema(schema)
        return self.validation_schema
